@@ -8,7 +8,7 @@
  */
 
 /*
- *  Copyright (C) 1997, 1998 Olivetti & Oracle Research Laboratory
+ *  Copyright (C) 1999 AT&T Laboratories Cambridge.  All Rights Reserved.
  *
  *  This is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -99,56 +99,57 @@ extern WindowPtr *WindowTable; /* Why isn't this in a header file? */
    to each client if there is one pending.  If the client is ready for it, we
    simply send the update.  If the client isn't ready we do nothing.  */
 
-#define SEND_FB_UPDATE(pScreen,prfb)			\
-  if (!prfb->dontSendFramebufferUpdate) {		\
-      rfbClientPtr cl;					\
-      for (cl = rfbClientHead; cl; cl = cl->next) {	\
-	  if (FB_UPDATE_PENDING(cl)) {			\
-	      if (cl->ready) {				\
-		  rfbSendFramebufferUpdate(cl);		\
-	      }						\
-	  }						\
-      }							\
+#define SEND_FB_UPDATE(pScreen,prfb)					\
+  if (!prfb->dontSendFramebufferUpdate) {				\
+      rfbClientPtr cl, nextCl;						\
+      for (cl = rfbClientHead; cl; cl = nextCl) {			\
+	  nextCl = cl->next;						\
+	  if (FB_UPDATE_PENDING(cl)) {					\
+	      if (REGION_NOTEMPTY(pScreen,&cl->requestedRegion)) {	\
+		  rfbSendFramebufferUpdate(cl);				\
+	      }								\
+	  }								\
+      }									\
   }
 
 
-static void PrintRegion _((ScreenPtr pScreen, RegionPtr reg));
-static void rfbCopyRegion _((ScreenPtr pScreen, rfbClientPtr cl,
-			     RegionPtr src, RegionPtr dst, int dx, int dy));
+static void PrintRegion(ScreenPtr pScreen, RegionPtr reg);
+static void rfbCopyRegion(ScreenPtr pScreen, rfbClientPtr cl,
+			  RegionPtr src, RegionPtr dst, int dx, int dy);
 
 /* GC funcs */
 
-static void rfbValidateGC _((GCPtr, unsigned long /*changes*/, DrawablePtr));
-static void rfbChangeGC _((GCPtr, unsigned long /*mask*/));
-static void rfbCopyGC _((GCPtr /*src*/, unsigned long /*mask*/, GCPtr /*dst*/));
-static void rfbDestroyGC _((GCPtr));
-static void rfbChangeClip _((GCPtr, int /*type*/, pointer /*pValue*/,
-			     int /*nrects*/));
-static void rfbDestroyClip _((GCPtr));
-static void rfbCopyClip _((GCPtr /*dst*/, GCPtr /*src*/));
+static void rfbValidateGC(GCPtr, unsigned long /*changes*/, DrawablePtr);
+static void rfbChangeGC(GCPtr, unsigned long /*mask*/);
+static void rfbCopyGC(GCPtr /*src*/, unsigned long /*mask*/, GCPtr /*dst*/);
+static void rfbDestroyGC(GCPtr);
+static void rfbChangeClip(GCPtr, int /*type*/, pointer /*pValue*/,
+			  int /*nrects*/);
+static void rfbDestroyClip(GCPtr);
+static void rfbCopyClip(GCPtr /*dst*/, GCPtr /*src*/);
 
 /* GC ops */
 
-static void rfbFillSpans _(());
-static void rfbSetSpans _(());
-static void rfbPutImage _(());
-static RegionPtr rfbCopyArea _(());
-static RegionPtr rfbCopyPlane _(());
-static void rfbPolyPoint _(());
-static void rfbPolylines _(());
-static void rfbPolySegment _(());
-static void rfbPolyRectangle _(());
-static void rfbPolyArc _(());
-static void rfbFillPolygon _(());
-static void rfbPolyFillRect _(());
-static void rfbPolyFillArc _(());
-static int rfbPolyText8 _(());
-static int rfbPolyText16 _(());
-static void rfbImageText8 _(());
-static void rfbImageText16 _(());
-static void rfbImageGlyphBlt _(());
-static void rfbPolyGlyphBlt _(());
-static void rfbPushPixels _(());
+static void rfbFillSpans();
+static void rfbSetSpans();
+static void rfbPutImage();
+static RegionPtr rfbCopyArea();
+static RegionPtr rfbCopyPlane();
+static void rfbPolyPoint();
+static void rfbPolylines();
+static void rfbPolySegment();
+static void rfbPolyRectangle();
+static void rfbPolyArc();
+static void rfbFillPolygon();
+static void rfbPolyFillRect();
+static void rfbPolyFillArc();
+static int rfbPolyText8();
+static int rfbPolyText16();
+static void rfbImageText8();
+static void rfbImageText16();
+static void rfbImageGlyphBlt();
+static void rfbPolyGlyphBlt();
+static void rfbPushPixels();
 
 
 static GCFuncs rfbGCFuncs = {
@@ -353,7 +354,7 @@ rfbClearToBackground (pWin, x, y, w, h, generateExposures)
 	box.x2 = w ? (box.x1 + w) : (pWin->drawable.x + pWin->drawable.width);
 	box.y2 = h ? (box.y1 + h) : (pWin->drawable.y + pWin->drawable.height);
 
-	REGION_INIT(pScreen, &tmpRegion, &box, 0);
+	SAFE_REGION_INIT(pScreen, &tmpRegion, &box, 0);
 
 	REGION_INTERSECT(pScreen, &tmpRegion, &tmpRegion, &pWin->clipList);
 
@@ -631,7 +632,7 @@ rfbPutImage(pDrawable, pGC, depth, x, y, w, h, leftPad, format, pBits)
     box.x2 = box.x1 + w;
     box.y2 = box.y1 + h;
 
-    REGION_INIT(pDrawable->pScreen, &tmpRegion, &box, 0);
+    SAFE_REGION_INIT(pDrawable->pScreen, &tmpRegion, &box, 0);
 
     REGION_INTERSECT(pDrawable->pScreen, &tmpRegion, &tmpRegion,
 		     WINDOW_CLIP_REGION((WindowPtr)pDrawable,pGC));
@@ -681,7 +682,7 @@ rfbCopyArea (pSrc, pDst, pGC, srcx, srcy, w, h, dstx, dsty)
     box.x2 = box.x1 + w;
     box.y2 = box.y1 + h;
 
-    REGION_INIT(pDst->pScreen, &dstRegion, &box, 0);
+    SAFE_REGION_INIT(pDst->pScreen, &dstRegion, &box, 0);
     REGION_INTERSECT(pDst->pScreen, &dstRegion, &dstRegion,
 		     WINDOW_CLIP_REGION((WindowPtr)pDst,pGC));
 
@@ -693,7 +694,7 @@ rfbCopyArea (pSrc, pDst, pGC, srcx, srcy, w, h, dstx, dsty)
 
 	for (cl = rfbClientHead; cl; cl = cl->next) {
 	    if (cl->useCopyRect) {
-		REGION_INIT(pSrc->pScreen, &srcRegion, &box, 0);
+		SAFE_REGION_INIT(pSrc->pScreen, &srcRegion, &box, 0);
 		REGION_INTERSECT(pSrc->pScreen, &srcRegion, &srcRegion,
 				 &((WindowPtr)pSrc)->clipList);
 
@@ -884,7 +885,7 @@ rfbCopyPlane (pSrc, pDst, pGC, srcx, srcy, w, h, dstx, dsty, plane)
     box.x2 = box.x1 + w;
     box.y2 = box.y1 + h;
 
-    REGION_INIT(pDst->pScreen, &tmpRegion, &box, 0);
+    SAFE_REGION_INIT(pDst->pScreen, &tmpRegion, &box, 0);
 
     REGION_INTERSECT(pDst->pScreen, &tmpRegion, &tmpRegion,
 		     WINDOW_CLIP_REGION((WindowPtr)pDst,pGC));
@@ -955,7 +956,7 @@ rfbPolyPoint (pDrawable, pGC, mode, npt, pts)
 	box.x2 = maxX + 1 + pDrawable->x;
 	box.y2 = maxY + 1 + pDrawable->y;
 
-	REGION_INIT(pDrawable->pScreen, &tmpRegion, &box, 0);
+	SAFE_REGION_INIT(pDrawable->pScreen, &tmpRegion, &box, 0);
 
 	REGION_INTERSECT(pDrawable->pScreen, &tmpRegion, &tmpRegion,
 			 WINDOW_CLIP_REGION(((WindowPtr)pDrawable),pGC));
@@ -1339,7 +1340,7 @@ rfbFillPolygon(pDrawable, pGC, shape, mode, count, pts)
 	box.x2 = maxX + 1 + pDrawable->x;
 	box.y2 = maxY + 1 + pDrawable->y;
 
-	REGION_INIT(pDrawable->pScreen, &tmpRegion, &box, 0);
+	SAFE_REGION_INIT(pDrawable->pScreen, &tmpRegion, &box, 0);
 
 	REGION_INTERSECT(pDrawable->pScreen, &tmpRegion, &tmpRegion,
 			 WINDOW_CLIP_REGION(((WindowPtr)pDrawable),pGC));
@@ -1527,7 +1528,7 @@ rfbPolyText8(pDrawable, pGC, x, y, count, chars)
     if (count) {
 	GetTextBoundingBox(pDrawable, pGC->font, x, y, count, &box);
 
-	REGION_INIT(pDrawable->pScreen, &tmpRegion, &box, 0);
+	SAFE_REGION_INIT(pDrawable->pScreen, &tmpRegion, &box, 0);
 
 	REGION_INTERSECT(pDrawable->pScreen, &tmpRegion, &tmpRegion,
 			 WINDOW_CLIP_REGION(((WindowPtr)pDrawable),pGC));
@@ -1569,7 +1570,7 @@ rfbPolyText16(pDrawable, pGC, x, y, count, chars)
     if (count) {
 	GetTextBoundingBox(pDrawable, pGC->font, x, y, count, &box);
 
-	REGION_INIT(pDrawable->pScreen, &tmpRegion, &box, 0);
+	SAFE_REGION_INIT(pDrawable->pScreen, &tmpRegion, &box, 0);
 
 	REGION_INTERSECT(pDrawable->pScreen, &tmpRegion, &tmpRegion,
 			 WINDOW_CLIP_REGION(((WindowPtr)pDrawable),pGC));
@@ -1610,7 +1611,7 @@ rfbImageText8(pDrawable, pGC, x, y, count, chars)
     if (count) {
 	GetTextBoundingBox(pDrawable, pGC->font, x, y, count, &box);
 
-	REGION_INIT(pDrawable->pScreen, &tmpRegion, &box, 0);
+	SAFE_REGION_INIT(pDrawable->pScreen, &tmpRegion, &box, 0);
 
 	REGION_INTERSECT(pDrawable->pScreen, &tmpRegion, &tmpRegion,
 			 WINDOW_CLIP_REGION(((WindowPtr)pDrawable),pGC));
@@ -1650,7 +1651,7 @@ rfbImageText16(pDrawable, pGC, x, y, count, chars)
     if (count) {
 	GetTextBoundingBox(pDrawable, pGC->font, x, y, count, &box);
 
-	REGION_INIT(pDrawable->pScreen, &tmpRegion, &box, 0);
+	SAFE_REGION_INIT(pDrawable->pScreen, &tmpRegion, &box, 0);
 
 	REGION_INTERSECT(pDrawable->pScreen, &tmpRegion, &tmpRegion,
 			 WINDOW_CLIP_REGION(((WindowPtr)pDrawable),pGC));
@@ -1691,7 +1692,7 @@ rfbImageGlyphBlt(pDrawable, pGC, x, y, nglyph, ppci, pglyphBase)
     if (nglyph) {
 	GetTextBoundingBox(pDrawable, pGC->font, x, y, nglyph, &box);
 
-	REGION_INIT(pDrawable->pScreen, &tmpRegion, &box, 0);
+	SAFE_REGION_INIT(pDrawable->pScreen, &tmpRegion, &box, 0);
 
 	REGION_INTERSECT(pDrawable->pScreen, &tmpRegion, &tmpRegion,
 			 WINDOW_CLIP_REGION(((WindowPtr)pDrawable),pGC));
@@ -1732,7 +1733,7 @@ rfbPolyGlyphBlt(pDrawable, pGC, x, y, nglyph, ppci, pglyphBase)
     if (nglyph) {
 	GetTextBoundingBox(pDrawable, pGC->font, x, y, nglyph, &box);
 
-	REGION_INIT(pDrawable->pScreen, &tmpRegion, &box, 0);
+	SAFE_REGION_INIT(pDrawable->pScreen, &tmpRegion, &box, 0);
 
 	REGION_INTERSECT(pDrawable->pScreen, &tmpRegion, &tmpRegion,
 			 WINDOW_CLIP_REGION(((WindowPtr)pDrawable),pGC));
@@ -1774,7 +1775,7 @@ rfbPushPixels(pGC, pBitMap, pDrawable, w, h, x, y)
     box.x2 = box.x1 + w;
     box.y2 = box.y1 + h;
 
-    REGION_INIT(pDrawable->pScreen, &tmpRegion, &box, 0);
+    SAFE_REGION_INIT(pDrawable->pScreen, &tmpRegion, &box, 0);
 
     REGION_INTERSECT(pDrawable->pScreen, &tmpRegion, &tmpRegion,
 		     WINDOW_CLIP_REGION(((WindowPtr)pDrawable),pGC));
@@ -1799,17 +1800,17 @@ PrintRegion(pScreen,reg)
     int nrects = REGION_NUM_RECTS(reg);
     int i;
 
-    fprintf(stderr,"Region num rects %d extents %d,%d %d,%d\n",nrects,
-	    (REGION_EXTENTS(pScreen,reg))->x1,
-	    (REGION_EXTENTS(pScreen,reg))->y1,
-	    (REGION_EXTENTS(pScreen,reg))->x2,
-	    (REGION_EXTENTS(pScreen,reg))->y2);
+    rfbLog("Region num rects %d extents %d,%d %d,%d\n",nrects,
+	   (REGION_EXTENTS(pScreen,reg))->x1,
+	   (REGION_EXTENTS(pScreen,reg))->y1,
+	   (REGION_EXTENTS(pScreen,reg))->x2,
+	   (REGION_EXTENTS(pScreen,reg))->y2);
 
     for (i = 0; i < nrects; i++) {
-	fprintf(stderr,"    rect %d,%d %dx%d\n",
-		REGION_RECTS(reg)[i].x1,
-		REGION_RECTS(reg)[i].y1,
-		REGION_RECTS(reg)[i].x2-REGION_RECTS(reg)[i].x1,
-		REGION_RECTS(reg)[i].y2-REGION_RECTS(reg)[i].y1);
+	rfbLog("    rect %d,%d %dx%d\n",
+	       REGION_RECTS(reg)[i].x1,
+	       REGION_RECTS(reg)[i].y1,
+	       REGION_RECTS(reg)[i].x2-REGION_RECTS(reg)[i].x1,
+	       REGION_RECTS(reg)[i].y2-REGION_RECTS(reg)[i].y1);
     }
 }
